package ru.dbotthepony.kstarbound.client.gl.program

import ru.dbotthepony.kstarbound.client.gl.BlendFunc
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.GLType
import ru.dbotthepony.kstarbound.client.gl.shader.GLShader
import ru.dbotthepony.kstarbound.client.gl.shader.GLTransformableColorableProgram
import ru.dbotthepony.kstarbound.client.gl.vertex.GLAttributeList
import ru.dbotthepony.kstarbound.client.gl.vertex.StreamVertexBuilder
import ru.dbotthepony.kstarbound.client.gl.vertex.GeometryType
import ru.dbotthepony.kstarbound.client.gl.vertex.QuadTransformers
import ru.dbotthepony.kstarbound.client.gl.vertex.quad
import ru.dbotthepony.kstarbound.client.render.LightRenderer
import ru.dbotthepony.kvector.vector.Color
import kotlin.properties.ReadOnlyProperty
import kotlin.reflect.KProperty

private class SimpleProgram<T : GLShaderProgram>(private val name: String, private val factory: (GLStateTracker, Array<out GLShader>) -> T) : ReadOnlyProperty<GLPrograms, T> {
	private var value: T? = null

	override fun getValue(thisRef: GLPrograms, property: KProperty<*>): T {
		thisRef.state.ensureSameThread()

		val value = value

		if (value != null) {
			return value
		}

		val vertex = GLShader.internalVertex("shaders/$name.vsh")
		val fragment = GLShader.internalFragment("shaders/$name.fsh")

		val newValue = factory.invoke(thisRef.state, arrayOf(vertex, fragment))

		this.value = newValue

		vertex.unlink()
		fragment.unlink()

		if (!newValue.linked) {
			newValue.link()
		}

		return newValue
	}
}

class GLLiquidProgram(state: GLStateTracker) : GLInternalProgram(state, "liquid") {
	init {
		link()
	}

	val baselineColor = this["baselineColor"]!!
	val transform = this["transform"]!!

	val builder by lazy {
		StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 16384)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).build()
	}
}

class GLLightProgram(state: GLStateTracker) : GLInternalProgram(state, "light") {
	init {
		link()
	}

	val baselineColor = this["baselineColor"]!!
	val transform = this["transform"]!!

	val builder by lazy {
		StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 32)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC2F).build()
	}
}

class GLHardLightGeometryProgram(state: GLStateTracker) : GLInternalProgram(state, listOf("hard_light_geometry"), listOf("hard_light_geometry"), listOf("hard_light_geometry")) {
	init {
		link()
	}

	val transform = this["transform"]!!
	val localToWorldTransform = this["localToWorldTransform"]!!
	val lightPosition = this["lightPosition"]!!

	val builder by lazy {
		StreamVertexBuilder(state, LightRenderer.SHADOW_FORMAT, GeometryType.QUADS_AS_LINES, 32)
	}
}

class GLSoftLightGeometryProgram(state: GLStateTracker) : GLInternalProgram(state, listOf("soft_light_geometry2"), listOf("soft_light_geometry2")) {
	init {
		link()
	}

	val transform = this["transform"]!!
	val lightPenetration = this["lightPenetration"]!!
	val localToWorldTransform = this["localToWorldTransform"]!!

	/**
	 * Vector3f(x, y, size)
	 */
	val lightPositionAndSize = this["lightPositionAndSize"]!!

	val builder by lazy {
		StreamVertexBuilder(state, LightRenderer.SHADOW_FORMAT_SOFT, GeometryType.QUADS_AS_LINES, 32)
	}
}

class GLColorQuadProgram(state: GLStateTracker) : GLInternalProgram(state, "screen_quad") {
	init {
		link()
	}

	val color = this["color"]!!

	private val builder by lazy {
		val builder = StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 1)

		builder.begin()
		builder.quad(-1f, -1f, 1f, 1f)
		builder.upload()

		builder
	}

	fun clearAlpha() {
		use()

		color.set(ALPHA)

		val old = state.blend
		val oldFunc = state.blendFunc

		state.blend = true
		state.blendFunc = BlendFunc.ONLY_ALPHA
		builder.draw()
		state.blend = old
		state.blendFunc = oldFunc
	}

	fun clearColor(color: Color = Color.WHITE) {
		use()

		this.color.set(color)

		val old = state.blend

		state.blend = false
		builder.draw()
		state.blend = old
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).build()
		val ALPHA = Color(0f, 0f, 0f, 1f)
	}
}

class GLTextureQuadProgram(state: GLStateTracker) : GLInternalProgram(state, "screen_quad_tex") {
	init {
		link()
	}

	val texture = this["texture0"]!!

	private val builder by lazy {
		val builder = StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 1)

		builder.begin()
		builder.quad(-1f, -1f, 1f, 1f, QuadTransformers.uv())
		builder.upload()

		builder
	}

	fun run(texture: Int = 0) {
		use()
		this.texture.set(texture)
		builder.draw()
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC2F).build()
	}
}

class GLFlatProgram(state: GLStateTracker, vararg shaders: GLShader) : GLTransformableColorableProgram(state, *shaders) {
	val builder by lazy {
		StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 1024)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).build()
	}
}

class GLPrograms(val state: GLStateTracker) {
	val tile by SimpleProgram("tile", ::GLTransformableColorableProgram)
	val font by SimpleProgram("font", ::GLTransformableColorableProgram)
	val flat by SimpleProgram("flat", ::GLFlatProgram)
	val liquid by lazy { GLLiquidProgram(state) }
	val light by lazy { GLLightProgram(state) }
	val hardLightGeometry by lazy { GLHardLightGeometryProgram(state) }
	val softLightGeometry by lazy { GLSoftLightGeometryProgram(state) }

	val colorQuad by lazy { GLColorQuadProgram(state) }
	val textureQuad by lazy { GLTextureQuadProgram(state) }
}
