package ru.dbotthepony.kstarbound.io

import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.RawPrototype
import ru.dbotthepony.kstarbound.defs.flattenJsonElement
import kotlin.reflect.KClass
import kotlin.reflect.KMutableProperty1
import kotlin.reflect.KType
import kotlin.reflect.full.isSuperclassOf

/**
 * Kotlin property aware adapter with arbitrary structure writer
 */
class ConfigurableTypeAdapter<T : RawPrototype<*, *>>(val factory: () -> T, vararg fields: KMutableProperty1<T, *>) : TypeAdapter<T>() {
	private val mappedFields = Object2ObjectArrayMap<String, KMutableProperty1<T, in Any?>>()
	// потому что returnType медленный
	private val mappedFieldsReturnTypes = Object2ObjectArrayMap<String, KType>()
	private val loggedMisses = ObjectArraySet<String>()

	init {
		for (field in fields) {
			// потому что в котлине нет понятия KProperty который не имеет getter'а, только setter
			require(mappedFields.put(field.name, field as KMutableProperty1<T, in Any?>) == null) { "${field.name} is defined twice" }
			mappedFieldsReturnTypes[field.name] = field.returnType
		}
	}

	val fields: Array<KMutableProperty1<T, in Any?>> get() {
		val iterator = mappedFields.values.iterator()
		return Array(mappedFields.size) { iterator.next() }
	}

	override fun write(writer: JsonWriter, value: T) {
		TODO("Not yet implemented")
	}

	override fun read(reader: JsonReader): T? {
		if (reader.peek() == JsonToken.NULL) {
			reader.nextNull()
			return null
		}

		reader.beginObject()
		val instance = factory.invoke()

		while (reader.hasNext()) {
			val name = reader.nextName()
			val field = mappedFields[name]

			if (field != null) {
				try {
					val peek = reader.peek()
					val expectedType = mappedFieldsReturnTypes[name]!!

					if (!expectedType.isMarkedNullable && peek == JsonToken.NULL) {
						throw IllegalArgumentException("Property ${field.name} of ${instance::class.qualifiedName} does not accept nulls")
					} else if (peek == JsonToken.NULL) {
						field.set(instance, null)
						reader.nextNull()
					} else {
						val classifier = expectedType.classifier

						if (classifier is KClass<*>) {
							if (classifier.isSuperclassOf(Float::class)) {
								val read = reader.nextDouble()
								instance.json[name] = read
								field.set(instance, read.toFloat())
							} else if (classifier.isSuperclassOf(Double::class)) {
								val read = reader.nextDouble()
								instance.json[name] = read
								field.set(instance, read)
							} else if (classifier.isSuperclassOf(Int::class)) {
								val read = reader.nextInt()
								instance.json[name] = read
								field.set(instance, read)
							} else if (classifier.isSuperclassOf(Long::class)) {
								val read = reader.nextLong()
								instance.json[name] = read
								field.set(instance, read)
							} else if (classifier.isSuperclassOf(String::class)) {
								val read = reader.nextString()
								instance.json[name] = read
								field.set(instance, read)
							} else if (classifier.isSuperclassOf(Boolean::class)) {
								val read = reader.nextBoolean()
								instance.json[name] = read
								field.set(instance, read)
							} else {
								val readElement = TypeAdapters.JSON_ELEMENT.read(reader)
								instance.json[name] = flattenJsonElement(readElement)
								field.set(instance, Starbound.gson.fromJson(readElement, classifier.java))
							}
						} else {
							throw IllegalStateException("Expected ${field.name} classifier to be KClass, got $classifier")
						}
					}
				} catch(err: Throwable) {
					throw JsonSyntaxException(
						"Reading property ${field.name} of ${instance::class.qualifiedName} near ${reader.path}",
						err
					)
				}
			} else {
				instance.json[name] = flattenJsonElement(TypeAdapters.JSON_ELEMENT.read(reader))

				if (!loggedMisses.contains(name)) {
					loggedMisses.add(name)
					LOGGER.warn("{} has no property for storing {}, this value will be visible to Lua scripts only", instance::class.qualifiedName, name)
				}
			}
		}

		reader.endObject()
		return instance
	}

	companion object {
		private val LOGGER = LogManager.getLogger(ConfigurableTypeAdapter::class.java)
	}
}